@@ -1,15 +1,33 @@
-"""Dataclasses that expose the metadata for the GetCapabilities call."""
+"""The configuration of a "feature type" in the WFS server.
+
+The "feature type" definitions define what models and attributes are exposed in the WFS server.
+When a model attribute is mentioned in the feature type, it can be exposed and queried against.
+Any field that is not mentioned in a definition, will therefore not be available, nor queryable.
+This metadata is used in the ``GetCapabilities`` call to advertise all available feature types.
+
+To handle other WFS request types besides ``GetCapabilities``, the "feature type" definition
+is translated internally into an internal XML Schema Definition (:mod:`gisserver.types`).
+That schema maps all model attributes to a specific XML layout, and includes
+all XSD Complex Types, elements and attributes linked to the Django model metadata.
+The feature type classes (and field types) offer a flexible translation
+from attribute listings into a schema definition.
+For example, model relationships can be modelled to a different XML layout.
+"""
+from __future__ import annotations
+
 import html
 import operator
+from collections import defaultdict
 from dataclasses import dataclass
 from functools import lru_cache, reduce
-from typing import List, Optional, Type, Union
+from typing import List, Union
 
+from django.conf import settings
 from django.contrib.gis.db import models as gis_models
 from django.contrib.gis.db.models import Extent, GeometryField
 from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
 from django.db import models
-from django.db.models.fields.reverse_related import ForeignObjectRel
+from django.db.models.fields.related import ForeignObjectRel  # Django 2.2 import
 from django.utils.functional import cached_property  # py3.8: functools
 
 from gisserver.db import conditional_transform
@@ -26,6 +44,11 @@ from gisserver.types import (
 )
 from gisserver.geometries import BoundingBox, CRS, WGS84
 
+if "django.contrib.postgres" in settings.INSTALLED_APPS:
+    from django.contrib.postgres.fields import ArrayField
+else:
+    ArrayField = None
+
 try:
     from typing import Literal  # Python 3.8
 
@@ -67,8 +90,15 @@ XSD_TYPES = {
 DEFAULT_XSD_TYPE = XsdTypes.anyType
 
 
-def get_basic_field_type(field_name: str, model_field: models.Field) -> XsdAnyType:
+def get_basic_field_type(
+    field_name: str, model_field: models.Field | ForeignObjectRel
+) -> XsdAnyType:
     """Determine the XSD field type for a Django field."""
+    if ArrayField is not None and isinstance(model_field, ArrayField):
+        # Determine the type based on the contents.
+        # The array notation is written as "is_many"
+        model_field = model_field.base_field
+
     try:
         # Direct instance, quickly resolved!
         return XSD_TYPES[model_field.__class__]
@@ -95,30 +125,22 @@ def get_basic_field_type(field_name: str, model_field: models.Field) -> XsdAnyTy
         return DEFAULT_XSD_TYPE
 
 
-def _get_target_model(model_field) -> Type[models.Model]:
-    """Find which model a related field points at"""
-    if isinstance(model_field, models.ForeignKey):
-        return model_field.target_field.model
-    elif isinstance(model_field, ForeignObjectRel):
-        # e.g. ManyToOneRel descriptor of a foreignkey_id field.
-        return model_field.remote_field.model
-    else:
-        raise NotImplementedError(
-            f"Model field is not supported as relation: {model_field.__class__.__name__}"
-        )
-
-
 def _get_model_fields(model, fields, parent=None):
     if fields == "__all__":
         # All regular fields
+        # Relationships will not be expanded since it can expose so many other fields.
+        # Only the relationships that store an ID on the model itself will be exposed.
         return [
             FeatureField(
-                name=f.attname if isinstance(f, models.ForeignKey) else f.name,
-                # .bind() is called directly:
+                name=f.attname,  # using attname so foreignkeys use the name_id field.
+                # .bind() is called directly by providing these arguments via init:
                 model=model,
                 parent=parent,
             )
             for f in model._meta.get_fields()
+            if not f.is_relation
+            or f.many_to_one  # ForeignKey
+            or f.one_to_one  # OneToOneField
         ]
     else:
         # Only defined fields
@@ -133,34 +155,35 @@ class ServiceDescription:
     """Basic metadata for an exposed GIS service."""
 
     title: str
-    abstract: Optional[str] = None
-    keywords: Optional[List[str]] = None
+    abstract: str | None = None
+    keywords: list[str] | None = None
 
-    provider_name: Optional[str] = None
-    provider_site: Optional[str] = None
-    contact_person: Optional[str] = None
+    provider_name: str | None = None
+    provider_site: str | None = None
+    contact_person: str | None = None
 
 
 class FeatureField:
-    """The configuration for an field inside a WFS Feature.
+    """The configuration for a field inside a WFS Feature.
 
     This defines how a Django model field is mapped into
     an XSD definition that the remaining application uses.
     """
 
     #: Allow to override the XSD element type that this field will generate.
-    xsd_element_class: Type[XsdElement] = XsdElement
+    xsd_element_class: type[XsdElement] = XsdElement
 
-    model: Optional[Type[models.Model]]
-    model_field: Optional[models.Field]
+    model: type[models.Model] | None
+    model_field: models.Field | ForeignObjectRel | None
 
     def __init__(
         self,
         name,
         model_attribute=None,
         model=None,
-        parent: "Optional[ComplexFeatureField]" = None,
+        parent: ComplexFeatureField | None = None,
         abstract=None,
+        xsd_class: type[XsdElement] | None = None,
     ):
         self.name = name
         self.model_attribute = model_attribute
@@ -169,6 +192,11 @@ class FeatureField:
         self.parent = parent
         self.abstract = abstract
 
+        # Allow to override the class attribute on 'self',
+        # which avoids having to subclass this field class as well.
+        if xsd_class is not None:
+            self.xsd_element_class = xsd_class
+
         self._nillable_relation = False
         if model is not None:
             self.bind(model)
@@ -181,11 +209,12 @@ class FeatureField:
 
     def bind(
         self,
-        model: Type[models.Model],
-        parent: "Optional[ComplexFeatureField]" = None,
+        model: type[models.Model],
+        parent: ComplexFeatureField | None = None,
     ):
         """Late-binding for the model.
 
+        This resolves the model field object from the provided model.
         This method is called internally when the field definition wasn't
         linked to a model yet. This allows the fields to be defined first,
         in external code, and then become part of the ``FeatureType`` fields
@@ -211,10 +240,20 @@ class FeatureField:
                     f" model_attribute: '{self.model_attribute}' can't be"
                     f" resolved for model '{self.model.__name__}'."
                 ) from e
+
             for name in field_path[1:]:
                 if field.null:
                     self._nillable_relation = True
-                field = _get_target_model(field)._meta.get_field(name)
+
+                if not field.is_relation:
+                    # Note this tests the previous loop variable, so it checks whether the
+                    # field can be walked into in order to resolve the next dotted name below.
+                    raise ImproperlyConfigured(
+                        f"FeatureField '{field.name}' has an invalid model_attribute: "
+                        f"field '{name}' is a '{field.__class__.__name__}', not a relation."
+                    )
+
+                field = field.related_model._meta.get_field(name)
             self.model_field = field
         else:
             self.model_field = self.model._meta.get_field(self.name)
@@ -230,12 +269,22 @@ class FeatureField:
         if self.model_field is None:
             raise RuntimeError(f"bind() was not called for {self!r}")
 
+        # Determine max number of occurrences.
+        if isinstance(self.model_field, GeometryField):
+            max_occurs = 1  # be explicit here, like mapserver does.
+        elif self.model_field.many_to_many or self.model_field.one_to_many:
+            max_occurs = "unbounded"  # M2M or reverse FK field
+        elif ArrayField is not None and isinstance(self.model_field, ArrayField):
+            max_occurs = self.model_field.size or "unbounded"
+        else:
+            max_occurs = None  # default is 1, but attribute can be left out.
+
         return self.xsd_element_class(
             name=self.name,
             type=self._get_xsd_type(),
             nillable=self.model_field.null or self._nillable_relation,
             min_occurs=0,
-            max_occurs=1 if isinstance(self.model_field, GeometryField) else None,
+            max_occurs=max_occurs,
             model_attribute=self.model_attribute,
             source=self.model_field,
         )
@@ -246,9 +295,11 @@ _FieldDefinitions = Union[_all_, List[_FieldDefinition]]
 
 
 class ComplexFeatureField(FeatureField):
-    """The configuration for an embedded foreign-key field.
+    """The configuration for an embedded relation field.
 
-    This translates the foreign key into an embedded XSD complex type.
+    This field type is suitable for any relational object, including
+    foreign keys, reverse relations and M2M fields. The internal logic
+    translates the relation into an embedded XSD complex type.
     """
 
     def __init__(
@@ -258,6 +309,8 @@ class ComplexFeatureField(FeatureField):
         model_attribute=None,
         model=None,
         abstract=None,
+        xsd_class=None,
+        xsd_base_type=XsdTypes.gmlAbstractFeatureType,
     ):
         """
         :param name: Name of the model field.
@@ -266,12 +319,17 @@ class ComplexFeatureField(FeatureField):
             Using ``__all__`` also works but is not recommended outside testing.
         """
         super().__init__(
-            name, model_attribute=model_attribute, model=model, abstract=abstract
+            name,
+            model_attribute=model_attribute,
+            model=model,
+            abstract=abstract,
+            xsd_class=xsd_class,
         )
+        self.xsd_base_type = xsd_base_type
         self._fields = fields
 
     @cached_property
-    def fields(self) -> List[FeatureField]:
+    def fields(self) -> list[FeatureField]:
         """Provide all fields that will be rendered as part of this complex field."""
         return _get_model_fields(self.target_model, self._fields, parent=self)
 
@@ -283,37 +341,38 @@ class ComplexFeatureField(FeatureField):
             name=f"{self.target_model._meta.object_name}Type",
             elements=[field.xsd_element for field in self.fields],
             attributes=[
-                # Add gml:id attribute definition so it can be resolved in xpath
+                # Add gml:id attribute definition, so it can be resolved in xpath
                 GmlIdAttribute(
                     type_name=self.name,
                     source=pk_field,
                     model_attribute=pk_field.name,
                 )
             ],
+            base=self.xsd_base_type,
             source=self.target_model,
         )
 
     @property
-    def target_model(self):
+    def target_model(self) -> type[models.Model]:
         """Detect which model the relation points to."""
         if self.model_field is None:
             raise RuntimeError("FeatureField.bind() is not called yet")
-
-        try:
-            return _get_target_model(self.model_field)
-        except NotImplementedError:
+        elif not self.model_field.is_relation:
             raise ImproperlyConfigured(
                 f"{self.__class__.__name__} does not support fields of "
                 f"type {self.model_field.__class__.__name__}."
-            ) from None
+            )
+        else:
+            return self.model_field.related_model
 
 
 def field(
     name: str,
     *,
     model_attribute=None,
-    abstract: Optional[str] = None,
-    fields: Optional[_FieldDefinitions] = None,
+    abstract: str | None = None,
+    fields: _FieldDefinitions | None = None,
+    xsd_class: type[XsdElement] | None = None,
 ) -> FeatureField:
     """Shortcut to define a WFS field.
 
@@ -327,10 +386,55 @@ def field(
     """
     if fields is not None:
         return ComplexFeatureField(
-            name=name, model_attribute=model_attribute, fields=fields, abstract=abstract
+            name=name,
+            model_attribute=model_attribute,
+            fields=fields,
+            abstract=abstract,
+            xsd_class=xsd_class,
         )
     else:
-        return FeatureField(name, model_attribute=model_attribute, abstract=abstract)
+        return FeatureField(
+            name,
+            model_attribute=model_attribute,
+            abstract=abstract,
+            xsd_class=xsd_class,
+        )
+
+
+@dataclass
+class FeatureRelation:
+    """Tell which related fields are queried by the feature.
+    Each dict holds an ORM-path, with the relevant sub-elements.
+    """
+
+    #: The ORM path that is queried for this particular relation
+    orm_path: str
+    #: The fields that will be retrieved for that path
+    orm_fields: set[str]
+    #: The model that is accessed for this relation (if set)
+    related_model: type[models.Model] | None
+    #: The source elements that access this relation.
+    xsd_elements: list[XsdElement]
+
+    @cached_property
+    def _local_model_field_names(self) -> list[str]:
+        """Tell which local fields of the model will be accessed by this feature."""
+
+        meta = self.related_model._meta
+        result = []
+        for name in self.orm_fields:
+            model_field = meta.get_field(name)
+            if not model_field.many_to_many and not model_field.one_to_many:
+                result.append(name)
+
+        # When this relation is retrieved through a ManyToOneRel (reverse FK),
+        # the prefetch_related() also needs to have the original foreign key
+        # in order to link all prefetches to the proper parent instance.
+        for xsd_element in self.xsd_elements:
+            if xsd_element.source is not None and xsd_element.source.one_to_many:
+                result.append(xsd_element.source.field.name)
+
+        return result
 
 
 class FeatureType:
@@ -344,23 +448,23 @@ class FeatureType:
     """
 
     #: Allow to override the XSD complex type that this feature will generate.
-    xsd_type_class: Type[XsdComplexType] = XsdComplexType
+    xsd_type_class: type[XsdComplexType] = XsdComplexType
 
     def __init__(
         self,
         queryset: models.QuerySet,
         *,
-        fields: Optional[_FieldDefinitions] = None,
+        fields: _FieldDefinitions | None = None,
         display_field_name: str = None,
         geometry_field_name: str = None,
         name: str = None,
         # WFS Metadata:
         title: str = None,
         abstract: str = None,
-        keywords: List[str] = None,
+        keywords: list[str] = None,
         crs: CRS = None,
-        other_crs: List[CRS] = None,
-        metadata_url: Optional[str] = None,
+        other_crs: list[CRS] = None,
+        metadata_url: str | None = None,
         # Settings
         show_name_field: bool = True,
         xml_prefix: str = "app",
@@ -411,12 +515,6 @@ class FeatureType:
         if html.escape(self.name) != self.name or " " in self.name or ":" in self.name:
             raise ValueError(f"Invalid feature name for XML: <{self.xml_name}>")
 
-        # Auto-detect geometry fields (also fills geometry_field_name)
-        self.geometry_fields = [
-            f
-            for f in self.model._meta.get_fields()
-            if isinstance(f, gis_models.GeometryField)
-        ]
         self._cached_resolver = lru_cache(100)(self._inner_resolve_element)
 
     def check_permissions(self, request):
@@ -431,25 +529,95 @@ class FeatureType:
         return f"{self.xml_prefix}:{self.name}"
 
     @cached_property
-    def supported_crs(self) -> List[CRS]:
+    def supported_crs(self) -> list[CRS]:
         """Return all spatial reference system ID's that this feature supports."""
         return [self.crs] + self.other_crs
 
     @cached_property
-    def geometry_field_names(self):
-        return {f.name for f in self.geometry_fields}
+    def geometry_fields(self) -> list[GeometryField]:
+        """Tell which fields of the model have a geometry field.
+        This only compares against fields that are mentioned in this feature type.
+        """
+        return [
+            ff.model_field or getattr(self.model, ff.model_attribute)
+            for ff in self.fields
+            if ff.xsd_element.is_geometry
+        ]
 
     @cached_property
-    def fields(self) -> List[FeatureField]:
+    def orm_relations(self) -> list[FeatureRelation]:
+        """Tell which fields will be retrieved from related fields.
+
+        This gives an object layout based on the XSD elements,
+        that can be used for prefetching data.
+        """
+        models = dict()
+        fields = defaultdict(set)
+        elements = defaultdict(list)
+
+        # Check all elements that render as "dotted" flattened relation
+        for xsd_element in self.xsd_type.flattened_elements:
+            if xsd_element.source is not None:
+                # Split "relation.field" notation into path, and take the field as child attribute.
+                obj_path, field = xsd_element.orm_relation
+                elements[obj_path].append(xsd_element)
+                fields[obj_path].add(field)
+                # field is already on relation:
+                models[obj_path] = xsd_element.source.model
+
+        # Check all elements that render as "nested" complex type:
+        for xsd_element in self.xsd_type.complex_elements:
+            # The complex element itself points to the root of the path,
+            # all sub elements become the child attributes.
+            obj_path = xsd_element.orm_path
+            elements[obj_path].append(xsd_element)
+            fields[obj_path] = {
+                f.orm_path
+                for f in xsd_element.type.elements
+                if not f.is_many or f.is_array  # exclude M2M, but include ArrayField
+            }
+            if xsd_element.source:
+                # field references a related object:
+                models[obj_path] = xsd_element.source.related_model
+
+        return [
+            FeatureRelation(
+                orm_path=obj_path,
+                orm_fields=sub_fields,
+                related_model=models.get(obj_path),
+                xsd_elements=elements[obj_path],
+            )
+            for obj_path, sub_fields in fields.items()
+        ]
+
+    @cached_property
+    def _local_model_field_names(self) -> list[str]:
+        """Tell which local fields of the model will be accessed by this feature."""
+        return [
+            (ff.model_field.name if ff.model_field else ff.model_attribute)
+            for ff in self.fields
+            if not ff.model_field.many_to_many and not ff.model_field.one_to_many
+        ]
+
+    @cached_property
+    def fields(self) -> list[FeatureField]:
         """Define which fields to render."""
         # This lazy reading allows providing 'fields' as lazy value.
         if self._fields is None:
-            return [FeatureField(self.geometry_field_name, model=self.model)]
+            # Autoconfig, no fields defined.
+            if not self._geometry_field_name:
+                self._geometry_field_name = next(
+                    f.name
+                    for f in self.model._meta.get_fields()
+                    if isinstance(f, gis_models.GeometryField)
+                )
+
+            return [FeatureField(self._geometry_field_name, model=self.model)]
         else:
             return _get_model_fields(self.model, self._fields)
 
     @cached_property
-    def display_field(self) -> Optional[models.Field]:
+    def display_field(self) -> models.Field | None:
         """Give access to the field that holds the string-representation."""
         if not self.show_name_field or not self.display_field_name:
             return None
@@ -461,24 +629,38 @@ class FeatureType:
         """Give access to the Django field that holds the geometry."""
         if not self.geometry_fields:
             raise ImproperlyConfigured(
-                f"Django model {self.model.__name__} does not have to a geometry field."
+                f"FeatureType '{self.name}' does not expose a geometry field."
             ) from None
 
-        if self.geometry_field_name:
-            return self.model._meta.get_field(self.geometry_field_name)
+        if self._geometry_field_name:
+            # Explicitly mentioned, use that field.
+            # Check whether the server is not accidentally exposing another field
+            # that is not part of the type definition.
+            field = next(
+                (
+                    field
+                    for field in self.geometry_fields
+                    if field.name == self._geometry_field_name
+                ),
+                None,
+            )
+            if field is None:
+                raise ImproperlyConfigured(
+                    f"FeatureType '{self.name}' does not expose the geometry field "
+                    f"'{self._geometry_field_name}' as part of its definition."
+                )
+
+            return field
         else:
+            # Default: take the first geometry
             return self.geometry_fields[0]
 
     @cached_property
     def geometry_field_name(self) -> str:
-        if self._geometry_field_name:
-            return self._geometry_field_name
-        else:
-            if not self.geometry_fields:
-                raise ImproperlyConfigured(
-                    f"Django model {self.model.__name__} does not have to a geometry field."
-                ) from None
-            return self.geometry_fields[0].name
+        """Tell which field is the geometry field."""
+        # Due to internal reorganization this property is no longer needed,
+        # retained for backward compatibility and to reflect any used input parameters.
+        return self.geometry_field.name
 
     @cached_property
     def crs(self) -> CRS:
@@ -491,16 +673,40 @@ class FeatureType:
 
     def get_queryset(self) -> models.QuerySet:
         """Return the queryset that is used as basis for this feature."""
-        # Adding .all() to avoid filling the caches.
-        return self.queryset.all()
+        # Return a queryset that only retrieves the fields that are actually displayed.
+        # That that without .only(), use at least `self.queryset.all()` so a clone is returned.
+        return self.queryset.only(*self._local_model_field_names)
 
-    def get_bounding_box(self) -> Optional[BoundingBox]:
+    def get_related_queryset(
+        self, feature_relation: FeatureRelation
+    ) -> models.QuerySet:
+        """Return the queryset that is used for prefetching related data."""
+        if feature_relation.related_model is None:
+            raise RuntimeError(
+                f"Unable to create prefetch queryset for relation {feature_relation.orm_path}, "
+                f"source model is not defined for: {feature_relation.xsd_elements!r}"
+            )
+        # Return a queryset that only retrieves the fields that are displayed.
+        return self.filter_related_queryset(
+            feature_relation.related_model.objects.only(
+                *feature_relation._local_model_field_names
+            )
+        )
+
+    def filter_related_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
+        """When a related object returns a queryset, this hook allows extra filtering."""
+        return queryset
+
+    def get_bounding_box(self) -> BoundingBox | None:
         """Returns a WGS84 BoundingBox for the complete feature.
 
         This is used by the GetCapabilities request. It may return ``None``
         when the database table is empty, or the custom queryset doesn't
         return any results.
         """
+        if not self.geometry_fields:
+            return None
+
         geo_expression = conditional_transform(
             self.geometry_field.name, self.geometry_field.srid, WGS84.srid
         )
@@ -508,9 +714,7 @@ class FeatureType:
         bbox = self.get_queryset().aggregate(a=Extent(geo_expression))["a"]
         return BoundingBox(*bbox, crs=WGS84) if bbox else None
 
-    def get_envelope(
-        self, instance, crs: Optional[CRS] = None
-    ) -> Optional[BoundingBox]:
+    def get_envelope(self, instance, crs: CRS | None = None) -> BoundingBox | None:
         """Get the bounding box for a single instance.
 
         This is only used for native Python rendering. When the database
@@ -551,8 +755,12 @@ class FeatureType:
         """
         pk_field = self.model._meta.pk
 
-        # Define <gml:boundedBy>
-        base_elements = [GmlBoundedByElement(feature_type=self)]
+        # Define <gml:boundedBy>, if the feature has a geometry
+        base_elements = []
+        if self.geometry_fields:
+            # Without a geometry, boundaries are not possible.
+            base_elements.append(GmlBoundedByElement(feature_type=self))
+
         if self.show_name_field:
             # Add <gml:name>
             gml_name = GmlNameElement(
@@ -589,6 +797,10 @@ class FeatureType:
 
     def resolve_element(self, xpath: str) -> XPathMatch:
         """Resolve the element, and the matching object.
+
+        This is used to convert XPath references in requests
+        to the actual elements and model attributes for queries.
+
         Internally, this method caches results.
         """
         nodes = self._cached_resolver(xpath)  # calls _inner_resolve_element
