@@ -3,11 +3,12 @@
 Note that the Django format_html() / mark_safe() logic is not used here,
 as it's quite a performance improvement to just use html.escape().
 """
+from __future__ import annotations
 import itertools
 import re
 from datetime import date, datetime, time
 from html import escape
-from typing import List, Optional, cast
+from typing import cast
 
 from django.contrib.gis import geos
 from django.db import models
@@ -24,15 +25,14 @@ from gisserver.db import (
     get_db_geometry_target,
 )
 from gisserver.exceptions import NotFound
-from gisserver.features import FeatureType
+from gisserver.features import FeatureRelation, FeatureType
 from gisserver.geometries import CRS
 from gisserver.parsers.fes20 import ValueReference
 from gisserver.types import XsdComplexType, XsdElement
 
-from .base import (
-    OutputRenderer,
-    StringBuffer,
-)
+from .base import OutputRenderer
+from .buffer import StringBuffer
+from .results import SimpleFeatureCollection
 
 GML_RENDER_FUNCTIONS = {}
 RE_GML_ID = re.compile(r'gml:id="[^"]+"')
@@ -161,8 +161,8 @@ class GML32Renderer(OutputRenderer):
                 self.start_collection(sub_collection)
                 if has_multiple_collections:
                     output.write(
-                        f"  <wfs:member>\n"
-                        f"    <wfs:{self.xml_collection_tag}"
+                        f"<wfs:member>\n"
+                        f"<wfs:{self.xml_collection_tag}"
                         f' timeStamp="{collection.timestamp}"'
                         f' numberMatched="{int(sub_collection.number_matched)}"'
                         f' numberReturned="{int(sub_collection.number_returned)}">\n'
@@ -176,18 +176,15 @@ class GML32Renderer(OutputRenderer):
                     # Only perform a 'yield' every once in a while,
                     # as it goes back-and-forth for writing it to the client.
                     if output.is_full():
-                        yield output.getvalue()
-                        output.clear()
+                        yield output.flush()
 
                 if has_multiple_collections:
-                    output.write(
-                        f"    </wfs:{self.xml_collection_tag}>\n  </wfs:member>\n"
-                    )
+                    output.write(f"</wfs:{self.xml_collection_tag}>\n</wfs:member>\n")
 
         output.write(f"</wfs:{self.xml_collection_tag}>\n")
-        yield output.getvalue()
+        yield output.flush()
 
-    def start_collection(self, sub_collection):
+    def start_collection(self, sub_collection: SimpleFeatureCollection):
         """Hook to allow initialization per feature type"""
         pass
 
@@ -198,7 +195,7 @@ class GML32Renderer(OutputRenderer):
         body = self.render_wfs_member_contents(
             feature_type, instance, extra_xmlns=extra_xmlns
         )
-        return f"  <wfs:member>\n{body}  </wfs:member>\n"
+        return f"<wfs:member>\n{body}</wfs:member>\n"
 
     def render_wfs_member_contents(
         self, feature_type: FeatureType, instance: models.Model, extra_xmlns=""
@@ -214,7 +211,7 @@ class GML32Renderer(OutputRenderer):
         pk = escape(str(instance.pk))
         output = StringBuffer()
         output.write(
-            f'    <{feature_type.xml_name} gml:id="{feature_type.name}.{pk}"{extra_xmlns}>\n'
+            f'<{feature_type.xml_name} gml:id="{feature_type.name}.{pk}"{extra_xmlns}>\n'
         )
 
         # Add all base class members, in their correct ordering
@@ -229,7 +226,7 @@ class GML32Renderer(OutputRenderer):
                     if gml is not None:
                         output.write(gml)
                 else:
-                    # e.g. <gml:name>
+                    # e.g. <gml:name>, or all other <app:...> nodes.
                     output.write(
                         self.render_element(feature_type, xsd_element, instance)
                     )
@@ -238,21 +235,19 @@ class GML32Renderer(OutputRenderer):
         for xsd_element in feature_type.xsd_type.elements:
             output.write(self.render_element(feature_type, xsd_element, instance))
 
-        output.write(f"    </{feature_type.xml_name}>\n")
+        output.write(f"</{feature_type.xml_name}>\n")
         return output.getvalue()
 
-    def render_bounds(self, feature_type, instance) -> Optional[str]:
+    def render_bounds(self, feature_type, instance) -> str | None:
         """Render the GML bounds for the complete instance"""
         envelope = feature_type.get_envelope(instance, self.output_crs)
         if envelope is not None:
             lower = " ".join(map(str, envelope.lower_corner))
             upper = " ".join(map(str, envelope.upper_corner))
-            return f"""      <gml:boundedBy>
-              <gml:Envelope srsDimension="2" srsName="{self.xml_srs_name}">
+            return f"""<gml:boundedBy><gml:Envelope srsDimension="2" srsName="{self.xml_srs_name}">
                 <gml:lowerCorner>{lower}</gml:lowerCorner>
                 <gml:upperCorner>{upper}</gml:upperCorner>
-              </gml:Envelope>
-            </gml:boundedBy>\n"""
+              </gml:Envelope></gml:boundedBy>\n"""
 
     def render_element(
         self, feature_type, xsd_element: XsdElement, instance: models.Model
@@ -275,20 +270,40 @@ class GML32Renderer(OutputRenderer):
         self, feature_type: FeatureType, xsd_element: XsdElement, value, extra_xmlns=""
     ) -> str:
         """Write the value of a single field."""
+        if xsd_element.is_many:
+            if value is None:
+                # No tag for optional element (see PropertyIsNull), otherwise xsi:nil node.
+                if xsd_element.min_occurs == 0:
+                    return ""
+                else:
+                    return f'<{xsd_element.xml_name} xsi:nil="true"{extra_xmlns}/>\n'
+            else:
+                # Render the tag multiple times
+                if xsd_element.type.is_complex_type:
+                    # If the retrieved QuerySet was not filtered yet, do so now. This can't
+                    # be done in get_value() because the FeatureType is not known there.
+                    value = feature_type.filter_related_queryset(value)
+
+                return "".join(
+                    self._render_xml_field(
+                        feature_type, xsd_element, value=item, extra_xmlns=extra_xmlns
+                    )
+                    for item in value
+                )
+        else:
+            return self._render_xml_field(
+                feature_type, xsd_element, value, extra_xmlns=extra_xmlns
+            )
+
+    def _render_xml_field(
+        self, feature_type: FeatureType, xsd_element: XsdElement, value, extra_xmlns=""
+    ) -> str:
         xml_name = xsd_element.xml_name
         if value is None:
-            return f'      <{xml_name} xsi:nil="true"{extra_xmlns} />\n'
+            return f'<{xml_name} xsi:nil="true"{extra_xmlns}/>\n'
         elif xsd_element.type.is_complex_type:
             # Expanded foreign relation / dictionary
-            xsd_type = cast(XsdComplexType, xsd_element.type)
-            output = StringBuffer()
-            output.write(f"      <{xml_name}>\n")
-            for sub_element in xsd_type.elements:
-                output.write(
-                    self.render_element(feature_type, sub_element, instance=value)
-                )
-            output.write(f"      </{xml_name}>\n")
-            return output.getvalue()
+            return self.render_xml_complex_type(feature_type, xsd_element, value)
         elif isinstance(value, datetime):
             value = value.astimezone(utc).isoformat()
         elif isinstance(value, (date, time)):
@@ -298,7 +313,17 @@ class GML32Renderer(OutputRenderer):
         else:
             value = escape(str(value))
 
-        return f"      <{xml_name}{extra_xmlns}>{value}</{xml_name}>\n"
+        return f"<{xml_name}{extra_xmlns}>{value}</{xml_name}>\n"
+
+    def render_xml_complex_type(self, feature_type, xsd_element, value) -> str:
+        """Write a single field, that consists of sub elements"""
+        xsd_type = cast(XsdComplexType, xsd_element.type)
+        output = StringBuffer()
+        output.write(f"<{xsd_element.xml_name}>\n")
+        for sub_element in xsd_type.elements:
+            output.write(self.render_element(feature_type, sub_element, instance=value))
+        output.write(f"</{xsd_element.xml_name}>\n")
+        return output.getvalue()
 
     def render_gml_field(
         self,
@@ -311,7 +336,7 @@ class GML32Renderer(OutputRenderer):
         """Write the value of an GML tag"""
         xml_name = xsd_element.xml_name
         gml = self.render_gml_value(value, gml_id=gml_id)
-        return f"      <{xml_name}{extra_xmlns}>{gml}</{xml_name}>\n"
+        return f"<{xml_name}{extra_xmlns}>{gml}</{xml_name}>\n"
 
     def render_gml_value(
         self, value: geos.GEOSGeometry, gml_id: str, extra_xmlns=""
@@ -426,32 +451,34 @@ class DBGML32Renderer(GML32Renderer):
             feature_type, queryset, output_crs, **params
         )
 
-        geometries = get_db_geometry_selects(
-            feature_type.xsd_type.geometry_elements, output_crs
-        )
-        return queryset.defer(*geometries.keys()).annotate(
-            _as_envelope_gml=cls.get_db_envelope_as_gml(
-                feature_type, queryset, output_crs
-            ),
-            **build_db_annotations(geometries, "_as_gml_{name}", AsGML),
-        )
+        # Retrieve geometries as pre-rendered instead.
+        gml_elements = feature_type.xsd_type.geometry_elements
+        geo_selects = get_db_geometry_selects(gml_elements, output_crs)
+        if geo_selects:
+            queryset = queryset.defer(*geo_selects.keys()).annotate(
+                _as_envelope_gml=cls.get_db_envelope_as_gml(
+                    feature_type, queryset, output_crs
+                ),
+                **build_db_annotations(geo_selects, "_as_gml_{name}", AsGML),
+            )
+
+        return queryset
 
     @classmethod
     def get_prefetch_queryset(
         cls,
         feature_type: FeatureType,
-        xsd_elements: List[XsdElement],
-        fields: List[str],
+        feature_relation: FeatureRelation,
         output_crs: CRS,
-    ):
+    ) -> models.QuerySet | None:
         """Perform DB annotations for prefetched relations too."""
-        model = xsd_elements[0].source
-        if model is None:
+        base = super().get_prefetch_queryset(feature_type, feature_relation, output_crs)
+        if base is None:
             return None
 
         # Find which fields are GML elements
         gml_elements = []
-        for e in xsd_elements:
+        for e in feature_relation.xsd_elements:
             if e.is_geometry:
                 # Prefetching a flattened relation
                 gml_elements.append(e)
@@ -461,11 +488,13 @@ class DBGML32Renderer(GML32Renderer):
                 gml_elements.extend(xsd_type.geometry_elements)
 
         geometries = get_db_geometry_selects(gml_elements, output_crs)
-        fields = set(fields) - set(geometries.keys())
         if geometries:
-            return model.objects.only("pk", *fields).annotate(
+            # Exclude geometries from the fields, fetch them as pre-rendered annotations instead.
+            return base.defer(geometries.keys()).annotate(
                 **build_db_annotations(geometries, "_as_gml_{name}", AsGML),
             )
+        else:
+            return base
 
     @classmethod
     def get_db_envelope_as_gml(cls, feature_type, queryset, output_crs) -> AsGML:
@@ -479,10 +508,17 @@ class DBGML32Renderer(GML32Renderer):
     @classmethod
     def _get_geometries_union(cls, feature_type: FeatureType, queryset, output_crs):
         """Combine all geometries of the model in a single SQL function."""
-        return conditional_transform(
-            get_geometries_union(feature_type.geometry_field_names, using=queryset.db),
-            feature_type.geometry_field.srid,
-            output_srid=output_crs.srid,
+        # Apply transforms where needed, in case some geometries use a different SRID.
+        return get_geometries_union(
+            [
+                conditional_transform(
+                    model_field.name,
+                    model_field.srid,
+                    output_srid=output_crs.srid,
+                )
+                for model_field in feature_type.geometry_fields
+            ],
+            using=queryset.db,
         )
 
     def render_element(
@@ -493,7 +529,7 @@ class DBGML32Renderer(GML32Renderer):
             value = get_db_annotation(instance, xsd_element.name, "_as_gml_{name}")
             if value is None:
                 # Avoid incrementing gml_seq
-                return f'      <{xsd_element.xml_name} xsi:nil="true" />\n'
+                return f'<{xsd_element.xml_name} xsi:nil="true"/>\n'
 
             self.gml_seq += 1
             return self.render_db_gml_field(
@@ -516,7 +552,7 @@ class DBGML32Renderer(GML32Renderer):
         """Write the value of an GML tag"""
         xml_name = xsd_element.xml_name
         if value is None:
-            return f'      <{xml_name} xsi:nil="true"{extra_xmlns} />\n'
+            return f'<{xml_name} xsi:nil="true"{extra_xmlns}/>\n'
 
         # Write the gml:id inside the first tag
         pos = value.find(">")
@@ -527,13 +563,13 @@ class DBGML32Renderer(GML32Renderer):
             first_tag += f' gml:id="{escape(gml_id)}"'
 
         gml = first_tag + value[pos:]
-        return f"      <{xml_name}{extra_xmlns}>{gml}</{xml_name}>\n"
+        return f"<{xml_name}{extra_xmlns}>{gml}</{xml_name}>\n"
 
     def render_bounds(self, feature_type, instance):
         """Generate the <gml:boundedBy> from DB prerendering."""
         gml = instance._as_envelope_gml
         if gml is not None:
-            return f"      <gml:boundedBy>{gml}</gml:boundedBy>\n"
+            return f"<gml:boundedBy>{gml}</gml:boundedBy>\n"
 
 
 class GML32ValueRenderer(GML32Renderer):
@@ -558,7 +594,7 @@ class GML32ValueRenderer(GML32Renderer):
         # Don't optimize queryset, it only retrieves one value
         return queryset
 
-    def start_collection(self, sub_collection):
+    def start_collection(self, sub_collection: SimpleFeatureCollection):
         # Resolve which XsdNode is being rendered
         match = sub_collection.feature_type.resolve_element(self.value_reference.xpath)
         self.xsd_node = match.child
@@ -568,18 +604,23 @@ class GML32ValueRenderer(GML32Renderer):
     ):
         """Overwritten to handle attribute support."""
         if self.xsd_node.is_attribute:
-            # Attributes are rendered without any spaces.
+            # When GetPropertyValue selects an attribute, it's value is rendered
+            # as plain-text (without spaces!) inside a <wfs:member> element.
             # The format_value() is needed for @gml:id
             body = self.xsd_node.format_value(instance["member"])
-            return f"  <wfs:member>{body}</wfs:member>\n"
+            return f"<wfs:member>{body}</wfs:member>\n"
         else:
+            # The call to GetPropertyValue selected an element.
+            # Render this single element tag inside the <wfs:member> parent.
             body = self.render_wfs_member_contents(feature_type, instance)
-            return f"  <wfs:member>\n{body}  </wfs:member>\n"
+            return f"<wfs:member>\n{body}</wfs:member>\n"
 
     def render_wfs_member_contents(
         self, feature_type: FeatureType, instance: dict, extra_xmlns=""
     ) -> str:
-        """Write the XML for a single object."""
+        """Write the XML for a single object.
+        In this case, it's only a single XML tag.
+        """
         value = instance["member"]
         if self.xsd_node.is_geometry:
             gml_id = self.get_gml_id(feature_type, instance["pk"], seq=1)
@@ -594,7 +635,7 @@ class GML32ValueRenderer(GML32Renderer):
             # The xsd_element is needed so render_xml_field() can render complex types.
             value = self.xsd_node.format_value(value)  # needed for @gml:id
             if self.xsd_node.is_attribute:
-                # For GetFeatureById, allow to return raw values
+                # For GetFeatureById, allow returning raw values
                 return str(value)
             else:
                 return self.render_xml_field(
@@ -635,7 +676,7 @@ class DBGML32ValueRenderer(DBGML32Renderer, GML32ValueRenderer):
                 instance["gml_member"],
                 gml_id=gml_id,
             )
-            return f"  <wfs:member>\n{body}  </wfs:member>\n"
+            return f"<wfs:member>\n{body}</wfs:member>\n"
         else:
             return super().render_wfs_member(
                 feature_type, instance, extra_xmlns=extra_xmlns
