@@ -1,11 +1,13 @@
+from __future__ import annotations
 from collections import deque
 
-from typing import List
+from typing import Iterable, cast
 
 from gisserver.features import FeatureType
 from gisserver.operations.base import WFSMethod
 from gisserver.types import XsdComplexType
-from .base import OutputRenderer, StringBuffer
+from .base import OutputRenderer
+from .buffer import StringBuffer
 
 
 class XMLSchemaRenderer(OutputRenderer):
@@ -16,7 +18,7 @@ class XMLSchemaRenderer(OutputRenderer):
 
     content_type = "application/gml+xml; version=3.2"  # mandatory for WFS
 
-    def __init__(self, method: WFSMethod, feature_types: List[FeatureType]):
+    def __init__(self, method: WFSMethod, feature_types: list[FeatureType]):
         """Overwritten method to handle the DescribeFeatureType context."""
         self.server_url = method.view.server_url
         self.app_xml_namespace = method.view.xml_namespace
@@ -28,7 +30,7 @@ class XMLSchemaRenderer(OutputRenderer):
         xmlns_features = "\n   ".join(
             f'xmlns:{p}="{self.app_xml_namespace}"'
             for p in sorted(
-                set(feature_type.xml_prefix for feature_type in self.feature_types)
+                {feature_type.xml_prefix for feature_type in self.feature_types}
             )
         )
 
@@ -99,21 +101,22 @@ class XMLSchemaRenderer(OutputRenderer):
         )
         return output.getvalue()
 
-    def _get_complex_types(self, root: XsdComplexType) -> List[XsdComplexType]:
-        """Find which fields of the XSDElements reference to complex types."""
-        worklist = deque(root.elements)
+    def _get_complex_types(self, root: XsdComplexType) -> Iterable[XsdComplexType]:
+        """Find all fields that reference to complex types, including nested elements."""
+        elements = deque(root.elements)
         complex_types = {}
 
         # Walk through next types, unless they are already seen.
         # No recursion is used here to handle circular loops of types.
-        while worklist:
-            element = worklist.popleft()
+        while elements:
+            element = elements.popleft()
             element_type = element.type
             if element_type.is_complex_type and element_type.name not in complex_types:
                 # ComplexType was not seen before, register it.
-                # It's members are added to the worklist.
-                complex_types[element_type.name] = element_type
-                worklist.extend(element_type.elements)
+                # It's members are added to the deque for recursive processing.
+                xsd_element_type = cast(XsdComplexType, element_type)
+                complex_types[element_type.name] = xsd_element_type
+                elements.extend(xsd_element_type.elements)
 
         # Present in a consistent order
         return complex_types.values()
