@@ -1,6 +1,7 @@
 """These classes map to the FES 2.0 specification for operators.
 The class names and attributes are identical to those in the FES spec.
 """
+from __future__ import annotations
 from itertools import groupby
 
 import operator
@@ -8,9 +9,10 @@ from dataclasses import dataclass, field
 from decimal import Decimal
 from enum import Enum
 from functools import reduce
-from typing import Any, Dict, List, Optional, Tuple, Union
-from xml.etree.ElementTree import Element
+from typing import Any, Union
+from xml.etree.ElementTree import Element, QName
 
+from django.conf import settings
 from django.contrib.gis import measure
 from django.db.models import Q
 from django.utils.functional import cached_property
@@ -18,7 +20,7 @@ from django.utils.functional import cached_property
 from gisserver.exceptions import ExternalParsingError, OperationProcessingFailed
 from gisserver.parsers import gml
 from gisserver.parsers.base import BaseNode, TagNameEnum, tag_registry
-from gisserver.parsers.utils import expect_tag, get_attribute, get_child
+from gisserver.parsers.tags import expect_children, expect_tag, get_attribute, get_child
 from gisserver.types import FES20
 from .identifiers import Id
 from .expressions import Expression, Literal, RhsTypes, ValueReference
@@ -40,6 +42,23 @@ else:
             ...
 
 
+if "django.contrib.postgres" in settings.INSTALLED_APPS:
+    # Comparisons with array fields go through a separate ORM lookup expression,
+    # so these can check whether ANY element matches in the array.
+    # This gives consistency between other repeated elements (e.g. M2M, reverse FK)
+    # where the whole object is returned when one of the sub-objects match.
+    ARRAY_LOOKUPS = {
+        "exact": "fes_anyexact",
+        "fes_notequal": "fes_anynotequal",
+        "lt": "fes_anylt",
+        "lte": "fes_anylte",
+        "gt": "fes_anygt",
+        "gte": "fes_anygte",
+    }
+else:
+    ARRAY_LOOKUPS = None
+
+
 class MatchAction(Enum):
     """Values for the 'matchAction' attribute of the BinaryComparisonOperator."""
 
@@ -91,7 +110,11 @@ class SpatialOperatorName(TagNameEnum):
 
 
 class TemporalOperatorName(TagNameEnum):
-    """XML tag names for datetime operators."""
+    """XML tag names for datetime operators.
+
+    Explanation here: http://old.geotools.org/Temporal-Filters_211091519.html
+    and: https://github.com/geotools/geotools/wiki/temporal-filters
+    """
 
     After = "after"
     Before = "before"
@@ -128,7 +151,7 @@ class Measure(BaseNode):
     xml_ns = FES20
 
     value: Decimal
-    uom: str  # Unit of measurement, Union[fes20:UomSymbol fes20:UomURI]
+    uom: str  # Unit of measurement, fes20:UomSymbol | fes20:UomURI
 
     @classmethod
     @expect_tag(FES20, "Distance")
@@ -144,7 +167,7 @@ class Operator(BaseNode):
 
     xml_ns = FES20
 
-    def build_query(self, compiler: CompiledQuery) -> Optional[Q]:
+    def build_query(self, compiler: CompiledQuery) -> Q | None:
         raise NotImplementedError(
             f"Using {self.__class__.__name__} is not supported yet."
         )
@@ -154,17 +177,17 @@ class Operator(BaseNode):
 class IdOperator(Operator):
     """List of ResourceId objects"""
 
-    id: List[Id]
+    id: list[Id]
 
     @property
-    def type_names(self) -> List[str]:
+    def type_names(self) -> list[str]:
         """Provide a list of all type names accessed by this operator"""
         return [
             type_name for type_name in self.grouped_ids.keys() if type_name is not None
         ]
 
     @cached_property
-    def grouped_ids(self) -> Dict[str, List[Id]]:
+    def grouped_ids(self) -> dict[str, list[Id]]:
         # For itertools.groupby(), items have to be sorted first.
         ids = sorted(self.id, key=operator.attrgetter("rid"))
         return {
@@ -192,7 +215,11 @@ class IdOperator(Operator):
 
 
 class NonIdOperator(Operator):
-    """Abstract base class, as defined by FES spec."""
+    """Abstract base class, as defined by FES spec.
+
+    This is used for nearly all operators,
+    except those that have <fes:ResourceId> elements as children.
+    """
 
     _source = None
     allow_geometries = False
@@ -202,7 +229,7 @@ class NonIdOperator(Operator):
         compiler: CompiledQuery,
         lhs: Expression,
         lookup: str,
-        rhs: Union[Expression, RhsTypes],
+        rhs: Expression | RhsTypes,
     ) -> Q:
         """Use the value in comparison with some other expression.
 
@@ -214,18 +241,23 @@ class NonIdOperator(Operator):
             lhs, rhs = rhs, lhs
 
         if compiler.feature_type is not None:
-            self.validate_comparison(compiler, lhs, lookup, rhs)
+            lookup = self.validate_comparison(compiler, lhs, lookup, rhs)
 
+        # Build Django Q-object
         lhs = lhs.build_lhs(compiler)
 
         if isinstance(rhs, (Expression, gml.GM_Object)):
             rhs = rhs.build_rhs(compiler)
 
-        result = Q(**{f"{lhs}__{lookup}": rhs})
-        return compiler.apply_extra_lookups(result)
+        comparison = Q(**{f"{lhs}__{lookup}": rhs})
+        return compiler.apply_extra_lookups(comparison)
 
     def validate_comparison(
-        self, compiler, lhs: Expression, lookup: str, rhs: Union[Expression, RhsTypes]
+        self,
+        compiler: CompiledQuery,
+        lhs: Expression,
+        lookup: str,
+        rhs: Expression | RhsTypes,
     ):
         """Validate whether a given comparison is even possible.
         Where needed, the lhs/rhs are already ordered in a logical sequence.
@@ -252,12 +284,28 @@ class NonIdOperator(Operator):
                 # can be validated before the ORM query is constructed.
                 xsd_element.validate_comparison(rhs.raw_value, lookup=lookup, tag=tag)
 
+            # Checking scalar values against array fields will fail.
+            # However, to make the queries consistent with other unbounded types (i.e. M2M fields),
+            # it makes sense to return an object when *one* entry in the array matches.
+            if xsd_element.is_array:
+                try:
+                    return ARRAY_LOOKUPS[lookup]
+                except KeyError:
+                    raise OperationProcessingFailed(
+                        "filter",
+                        f"Operator '{tag}' is not supported for "
+                        f"the '{xsd_element.name}' property.",
+                        status_code=400,  # not HTTP 500 here. Spec allows both.
+                    ) from None
+
+        return lookup
+
     def build_compare_between(
         self,
         compiler: CompiledQuery,
         lhs: Expression,
         lookup: str,
-        rhs: Tuple[HasBuildRhs, HasBuildRhs],
+        rhs: tuple[HasBuildRhs, HasBuildRhs],
     ) -> Q:
         """Use the value in comparison with 2 other values (e.g. between query)"""
         if compiler.feature_type is not None:
@@ -265,7 +313,7 @@ class NonIdOperator(Operator):
             self.validate_comparison(compiler, lhs, lookup, rhs[1])
 
         field_name = lhs.build_lhs(compiler)
-        result = Q(
+        comparison = Q(
             **{
                 f"{field_name}__{lookup}": (
                     rhs[0].build_rhs(compiler),
@@ -273,7 +321,7 @@ class NonIdOperator(Operator):
                 )
             }
         )
-        return compiler.apply_extra_lookups(result)
+        return compiler.apply_extra_lookups(comparison)
 
 
 class SpatialOperator(NonIdOperator):
@@ -291,9 +339,10 @@ class DistanceOperator(SpatialOperator):
     operatorType: DistanceOperatorName
     geometry: gml.GM_Object
     distance: Measure
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(3, ValueReference, gml.GM_Object, Measure)
     def from_xml(cls, element: Element):
         geometries = gml.find_gml_nodes(element)
         if not geometries:
@@ -330,9 +379,9 @@ class BinarySpatialOperator(SpatialOperator):
     allow_geometries = True  # override static attribute
 
     operatorType: SpatialOperatorName
-    operand1: Optional[ValueReference]
+    operand1: ValueReference | None
     operand2: SpatialDescription
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
     def from_xml(cls, element: Element):
@@ -358,7 +407,7 @@ class BinarySpatialOperator(SpatialOperator):
     def build_query(self, compiler: CompiledQuery) -> Q:
         operant1 = self.operand1
         if operant1 is None:
-            operant1 = ValueReference(xpath=compiler.feature_type.geometry_field_name)
+            operant1 = ValueReference(xpath=compiler.feature_type.geometry_field.name)
 
         return self.build_compare(
             compiler,
@@ -376,9 +425,10 @@ class TemporalOperator(NonIdOperator):
     operatorType: TemporalOperatorName
     operand1: ValueReference
     operand2: TemporalOperand
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(2, ValueReference, TemporalOperand)
     def from_xml(cls, element: Element):
         return cls(
             operatorType=TemporalOperatorName.from_xml(element),
@@ -404,12 +454,13 @@ class BinaryComparisonOperator(ComparisonOperator):
     """A comparison between 2 values, e.g. A == B."""
 
     operatorType: BinaryComparisonName
-    expression: Tuple[Expression, Expression]
+    expression: tuple[Expression, Expression]
     matchCase: bool = True
     matchAction: MatchAction = MatchAction.Any
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(2, Expression, Expression)
     def from_xml(cls, element: Element):
         return cls(
             operatorType=BinaryComparisonName.from_xml(element),
@@ -439,12 +490,31 @@ class BetweenComparisonOperator(ComparisonOperator):
     expression: Expression
     lowerBoundary: Expression
     upperBoundary: Expression
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(3, Expression, "LowerBoundary", "UpperBoundary")
     def from_xml(cls, element: Element):
+        if element[1].tag != QName(FES20, "LowerBoundary") or element[2].tag != QName(
+            FES20, "UpperBoundary"
+        ):
+            raise ExternalParsingError(
+                f"{element.tag} should have 3 child nodes: "
+                f"(expression), <LowerBoundary>, <UpperBoundary>"
+            )
+
         lower = get_child(element, FES20, "LowerBoundary")
         upper = get_child(element, FES20, "UpperBoundary")
+
+        if len(lower) != 1:
+            raise ExternalParsingError(
+                f"{lower.tag} should have 1 expression child node"
+            )
+        if len(upper) != 1:
+            raise ExternalParsingError(
+                f"{upper.tag} should have 1 expression child node"
+            )
+
         return cls(
             expression=Expression.from_child_xml(element[0]),
             lowerBoundary=Expression.from_child_xml(lower[0]),
@@ -466,19 +536,21 @@ class BetweenComparisonOperator(ComparisonOperator):
 class LikeOperator(ComparisonOperator):
     """Perform wildcard matching."""
 
-    expression: Tuple[Expression, Expression]
+    expression: tuple[Expression, Expression]
     wildCard: str
     singleChar: str
     escapeChar: str
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(2, Expression, Expression)
     def from_xml(cls, element: Element):
         return cls(
             expression=(
                 Expression.from_child_xml(element[0]),
                 Expression.from_child_xml(element[1]),
             ),
+            # These attributes are required by the WFS spec:
             wildCard=get_attribute(element, "wildCard"),
             singleChar=get_attribute(element, "singleChar"),
             escapeChar=get_attribute(element, "escapeChar"),
@@ -512,14 +584,15 @@ class LikeOperator(ComparisonOperator):
 @tag_registry.register("PropertyIsNil")
 class NilOperator(ComparisonOperator):
     """Check whether the value evaluates to null/None.
-    If the WFS would return a property element with <tns:p xsi:nil='true'>, this returns true.
+    If the WFS returned a property element with <tns:p xsi:nil='true'>, this returns true.
     """
 
-    expression: Optional[Expression]
+    expression: Expression | None
     nilReason: str
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(1, Expression)
     def from_xml(cls, element: Element):
         return cls(
             expression=Expression.from_child_xml(element[0]) if element else None,
@@ -542,9 +615,10 @@ class NullOperator(ComparisonOperator):
     """
 
     expression: Expression
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(1, Expression)
     def from_xml(cls, element: Element):
         return cls(
             expression=Expression.from_child_xml(element[0]), _source=element.tag
@@ -554,6 +628,7 @@ class NullOperator(ComparisonOperator):
         # For now, the implementation is identical to PropertyIsNil.
         # According to the WFS spec, this should only be true when the element
         # is not returned at all (minOccurs=0).
+        # TODO: this happens for maxOccurs=unbounded with a null value.
         return self.build_compare(
             compiler, lhs=self.expression, lookup="isnull", rhs=True
         )
@@ -569,11 +644,12 @@ class LogicalOperator(NonIdOperator):
 class BinaryLogicOperator(LogicalOperator):
     """Apply an AND or OR operator"""
 
-    operands: List[NonIdOperator]
+    operands: list[NonIdOperator]
     operatorType: BinaryLogicType
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(2, NonIdOperator, NonIdOperator)
     def from_xml(cls, element: Element):
         return cls(
             operands=[NonIdOperator.from_child_xml(child) for child in element],
@@ -594,9 +670,10 @@ class UnaryLogicOperator(LogicalOperator):
 
     operands: NonIdOperator
     operatorType: UnaryLogicType
-    _source: Optional[str] = field(compare=False, default=None)
+    _source: str | None = field(compare=False, default=None)
 
     @classmethod
+    @expect_children(1, NonIdOperator)
     def from_xml(cls, element: Element):
         return cls(
             operands=NonIdOperator.from_child_xml(element[0]),
